from .api import CDFResourceAPI, Endpoint
from .responses import PagedResponse, QueryResponse, ResponseItems

__all__ = [
    "CDFResourceAPI",
    "Endpoint",
    "PagedResponse",
    "QueryResponse",
    "ResponseItems",
]



from typing import Any, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .filemetadata import FileMetadataRequest
from .identifiers import ExternalId


class StreamlitFile(BaseModelObject):
    """Base class for Streamlit app with common fields."""

    external_id: str
    name: str
    creator: str
    entrypoint: str | None = None
    description: str | None = None
    published: bool = False
    theme: Literal["Light", "Dark"] = "Light"
    thumbnail: str | None = None
    data_set_id: int | None = None
    cognite_toolkit_app_hash: str | None = None


class StreamlitRequest(StreamlitFile, UpdatableRequestResource):
    """Request resource for creating/updating Streamlit apps."""

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    def as_file(self) -> FileMetadataRequest:
        metadata: dict[str, str] = {
            "creator": self.creator,
            "name": self.name,
            "published": str(self.published),
            "theme": str(self.theme),
        }
        if self.description:
            metadata["description"] = self.description
        if self.thumbnail:
            metadata["thumbnail"] = self.thumbnail
        if self.cognite_toolkit_app_hash:
            metadata["cdf-toolkit-app-hash"] = self.cognite_toolkit_app_hash
        if self.entrypoint:
            metadata["entrypoint"] = self.entrypoint
        return FileMetadataRequest(
            external_id=self.external_id,
            name=f"{self.name}-source.json",
            data_set_id=self.data_set_id,
            directory="/streamlit-apps/",
            metadata=metadata,
        )

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        return self.as_file().dump(camel_case=camel_case, exclude_extra=exclude_extra)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        return self.as_file().as_update(mode)


class StreamlitResponse(StreamlitFile, ResponseResource[StreamlitRequest]):
    """Response resource for Streamlit apps."""

    created_time: int
    last_updated_time: int

    def as_request_resource(self) -> StreamlitRequest:
        return StreamlitRequest.model_validate(self.dump(), extra="ignore")



import sys
from typing import Any

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawDatabaseId, RawTableId

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class RAWDatabaseRequest(RequestResource):
    name: str = Field(alias="dbName")

    def as_id(self) -> RawDatabaseId:
        return RawDatabaseId(name=self.name)

    # Override dump to always use by_alias=False since the API expects name='...'
    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        if exclude_extra:
            return self.model_dump(
                mode="json",
                by_alias=False,
                exclude_unset=True,
                exclude=set(self.__pydantic_extra__) if self.__pydantic_extra__ else None,
            )
        return self.model_dump(mode="json", by_alias=False, exclude_unset=True)


class RAWDatabaseResponse(ResponseResource[RAWDatabaseRequest]):
    name: str
    created_time: int

    def as_request_resource(self) -> RAWDatabaseRequest:
        return RAWDatabaseRequest.model_validate(self.dump(), extra="ignore")

    def as_id(self) -> RawDatabaseId:
        return RawDatabaseId(name=self.name)


class RAWTableRequest(RequestResource):
    # This is a query parameter, so we exclude it from serialization.
    db_name: str = Field(exclude=True)
    name: str = Field(alias="tableName")

    def as_id(self) -> RawTableId:
        return RawTableId(db_name=self.db_name, name=self.name)

    # Override dump to always use by_alias=False since the API expects name='...'
    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        if exclude_extra:
            return self.model_dump(
                mode="json",
                by_alias=False,
                exclude_unset=True,
                exclude=set(self.__pydantic_extra__) if self.__pydantic_extra__ else None,
            )
        return self.model_dump(mode="json", by_alias=False, exclude_unset=True)


class RAWTableResponse(ResponseResource[RAWTableRequest]):
    # This is a query parameter, so we exclude it from serialization.
    # Default to empty string to allow parsing from API responses (which don't include db_name).
    db_name: str = Field(default="", exclude=True)
    name: str
    created_time: int

    def as_request_resource(self) -> RAWTableRequest:
        dumped = {**self.dump(), "dbName": self.db_name}
        return RAWTableRequest.model_validate(dumped, extra="ignore")

    @classmethod
    def _load(cls, resource: dict[str, Any]) -> Self:
        """Load method to match CogniteResource signature."""
        return cls.model_validate(resource, by_name=True)

    def as_id(self) -> RawTableId:
        return RawTableId(db_name=self.db_name, name=self.name)



from abc import ABC

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from ._references import SpaceReference


class Space(BaseModelObject, ABC):
    space: str
    name: str | None = None
    description: str | None = None

    def as_id(self) -> SpaceReference:
        return SpaceReference(space=self.space)


class SpaceRequest(Space, RequestResource): ...


class SpaceResponse(Space, ResponseResource[SpaceRequest]):
    created_time: int
    last_updated_time: int
    is_global: bool

    def as_request_resource(self) -> SpaceRequest:
        return SpaceRequest.model_validate(self.model_dump(by_alias=True), extra="ignore")



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawDatabaseId, RawTableId
from cognite_toolkit._cdf_tk.client.resource_classes.raw import (
    RAWDatabaseRequest,
    RAWDatabaseResponse,
    RAWTableRequest,
    RAWTableResponse,
)


class RawDatabasesAPI(CDFResourceAPI[RawDatabaseId, RAWDatabaseRequest, RAWDatabaseResponse]):
    """API for managing RAW databases in CDF.

    This API provides methods to create, list, and delete RAW databases.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client,
            {
                "create": Endpoint(method="POST", path="/raw/dbs", item_limit=1000),
                "delete": Endpoint(method="POST", path="/raw/dbs/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/raw/dbs", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RAWDatabaseResponse]:
        return PagedResponse[RAWDatabaseResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[RAWDatabaseResponse]:
        return ResponseItems[RAWDatabaseResponse].model_validate_json(response.body)

    def create(self, items: Sequence[RAWDatabaseRequest]) -> list[RAWDatabaseResponse]:
        """Create databases in CDF.

        Args:
            items: List of RAWDatabase objects to create.

        Returns:
            List of created RAWDatabase objects.
        """
        return self._request_item_response(list(items), "create")

    def delete(self, items: Sequence[RawDatabaseId], recursive: bool = False) -> None:
        """Delete databases from CDF.

        Args:
            items: List of RAWDatabase objects to delete.
            recursive: Whether to delete tables within the database recursively.
        """
        self._request_no_response(items, "delete", extra_body={"recursive": recursive})

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RAWDatabaseResponse]:
        """Iterate over all databases in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RAWDatabase objects.
        """
        return self._paginate(limit=limit, cursor=cursor)

    def iterate(
        self,
        limit: int = 100,
    ) -> Iterable[list[RAWDatabaseResponse]]:
        """Iterate over all databases in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of RAWDatabase objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = None) -> list[RAWDatabaseResponse]:
        """List all databases in CDF.

        Args:
            limit: Maximum number of databases to return. If None, returns all databases.

        Returns:
            List of RAWDatabase objects.
        """
        return self._list(limit=limit)


class RawTablesAPI(CDFResourceAPI[RawTableId, RAWTableRequest, RAWTableResponse]):
    """API for managing RAW tables in CDF.

    This API provides methods to create, list, and delete RAW tables within a database.

    Note: This API requires db_name as a path parameter for all operations,
    so it overrides several base class methods to handle dynamic paths.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        # We pass empty endpoint map since paths are dynamic (depend on db_name)
        super().__init__(
            http_client,
            {
                "create": Endpoint(method="POST", path="/raw/dbs/{dbName}/tables", item_limit=1000),
                "delete": Endpoint(method="POST", path="/raw/dbs/{dbName}/tables/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/raw/dbs/{dbName}/tables", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RAWTableResponse]:
        """Parse a page response. Note: db_name must be injected separately."""
        return PagedResponse[RAWTableResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[RAWTableResponse]:
        """Parse a reference response. Note: db_name must be injected separately."""
        return ResponseItems[RAWTableResponse].model_validate_json(response.body)

    def create(self, items: Sequence[RAWTableRequest], ensure_parent: bool = False) -> list[RAWTableResponse]:
        """Create tables in a database in CDF.

        Args:
            items: List of RAWTable objects to create.
            ensure_parent: Create database if it doesn't exist already.

        Returns:
            List of created RAWTable objects.
        """
        result: list[RAWTableResponse] = []
        for (db_name,), group in self._group_items_by_text_field(items, "db_name").items():
            if not db_name:
                raise ValueError("db_name must be set on all RAWTable items for creation.")
            endpoint = f"/raw/dbs/{db_name}/tables"
            created = self._request_item_response(
                group, "create", params={"ensureParent": ensure_parent}, endpoint=endpoint
            )
            for table in created:
                table.db_name = db_name
                result.append(table)
        return result

    def delete(self, items: Sequence[RawTableId]) -> None:
        """Delete tables from a database in CDF.

        Args:
            items: List of RAWTable objects to delete.
        """
        for (db_name,), group in self._group_items_by_text_field(items, "db_name").items():
            if not db_name:
                raise ValueError("db_name must be set on all RAWTable items for deletion.")
            endpoint = f"/raw/dbs/{db_name}/tables/delete"
            self._request_no_response(list(group), "delete", endpoint=endpoint)

    def paginate(
        self,
        db_name: str,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RAWTableResponse]:
        """Iterate over all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RAWTable objects.
        """
        page = self._paginate(cursor=cursor, limit=limit, endpoint_path=f"/raw/dbs/{db_name}/tables")
        for table in page.items:
            table.db_name = db_name
        return page

    def iterate(
        self,
        db_name: str,
        limit: int = 100,
    ) -> Iterable[list[RAWTableResponse]]:
        """Iterate over all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of RAWTable objects.
        """
        for table in self._iterate(limit=limit, endpoint_path=f"/raw/dbs/{db_name}/tables"):
            for t in table:
                t.db_name = db_name
            yield table

    def list(self, db_name: str, limit: int | None = None) -> list[RAWTableResponse]:
        """List all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of tables to return. If None, returns all tables.

        Returns:
            List of RAWTable objects.
        """
        listed = self._list(limit, endpoint_path=f"/raw/dbs/{db_name}/tables")
        for table in listed:
            table.db_name = db_name
        return listed


class RawAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self.databases = RawDatabasesAPI(http_client)
        self.tables = RawTablesAPI(http_client)



from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field, JsonValue

from cognite_toolkit._cdf_tk.client.resource_classes.graphql_data_model import GraphQLDataModelResponse

T = TypeVar("T", bound=BaseModel)


class ResponseItems(BaseModel, Generic[T]):
    """A page of reference items from a paginated API response.

    Attributes:
        items: The list of reference items in this page.
    """

    items: list[T]


class PagedResponse(BaseModel, Generic[T]):
    items: list[T]
    next_cursor: str | None = Field(None, alias="nextCursor")


class QueryResponse(BaseModel):
    items: dict[str, list[dict[str, Any]]]
    typing: dict[str, JsonValue] | None = None
    next_cursor: dict[str, str] = Field(alias="nextCursor")
    debug: dict[str, JsonValue] | None = None


class GraphQLResponse(BaseModel):
    data: GraphQLDataModelResponse
    errors: list[dict[str, JsonValue]] | None = None


class GraphQLUpsertResponse(BaseModel):
    upsert_graph_ql_dml_version: GraphQLResponse = Field(alias="upsertGraphQlDmlVersion")



from ._build import BuildFiles, BuildFolder, BuildParameters, BuiltModule
from ._config import ConfigYAML
from ._insights import ConsistencyError, Insight, InsightList, ModelSyntaxError, Recommendation
from ._module import Module, ModuleSource, ResourceType
from ._types import AbsoluteDirPath, RelativeDirPath, RelativeFilePath, ValidationType

__all__ = [
    "AbsoluteDirPath",
    "BuildFiles",
    "BuildFolder",
    "BuildParameters",
    "BuiltModule",
    "ConfigYAML",
    "ConsistencyError",
    "Insight",
    "InsightList",
    "ModelSyntaxError",
    "Module",
    "ModuleSource",
    "Recommendation",
    "RelativeDirPath",
    "RelativeFilePath",
    "ResourceType",
    "ValidationType",
]



from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject


class UserProfilesConfiguration(BaseModelObject):
    enabled: bool


class Claim(BaseModelObject):
    claim_name: str


class OidcConfiguration(BaseModelObject):
    jwks_url: str
    token_url: str | None = None
    issuer: str
    audience: str
    skew_ms: int | None = None
    access_claims: list[Claim]
    scope_claims: list[Claim]
    log_claims: list[Claim]
    is_group_callback_enabled: bool | None = None
    identity_provider_scope: str | None = None


class OrganizationResponse(BaseModelObject):
    name: str
    url_name: str
    organization: str
    user_profiles_configuration: UserProfilesConfiguration
    oidc_configuration: OidcConfiguration | None = None



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, TypeAdapter, model_serializer
from pydantic_core.core_schema import FieldSerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject


class IndexDefinition(BaseModelObject, ABC):
    index_type: str
    properties: list[str]

    @model_serializer(mode="wrap")  # type: ignore[type-var]
    def serialize(self, handler: SerializerFunctionWrapHandler, info: FieldSerializationInfo) -> dict[str, Any]:
        # Always serialize as {"indexType": self.index_type}, even if model_dump(exclude_unset=True)
        serialized = handler(self)
        return {"indexType" if info.by_alias else "index_type": self.index_type, **serialized}


class BtreeIndex(IndexDefinition):
    index_type: Literal["btree"] = "btree"
    by_space: bool | None = None
    cursorable: bool | None = None


class InvertedIndex(IndexDefinition):
    index_type: Literal["inverted"] = "inverted"


Index = Annotated[BtreeIndex | InvertedIndex, Field(discriminator="index_type")]

IndexAdapter: TypeAdapter[Index] = TypeAdapter(Index)



from collections.abc import Iterable, Mapping, Sequence
from types import MappingProxyType
from typing import ClassVar, cast

from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling import EdgeId, NodeId, ViewId
from cognite.client.utils._identifier import InstanceId

from cognite_toolkit._cdf_tk import constants
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.request_classes.filters import InstanceFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ContainerReference,
    InstanceRequest,
    InstanceResponse,
    SpaceReference,
    ViewReference,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceRequestAdapter
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    TypedViewReference,
)
from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, SpaceCRUD, ViewCRUD
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from . import StorageIOConfig
from ._base import ConfigurableStorageIO, Page, UploadableStorageIO
from .selectors import InstanceFileSelector, InstanceSelector, InstanceSpaceSelector, InstanceViewSelector, SelectedView


class InstanceIO(
    ConfigurableStorageIO[InstanceSelector, InstanceResponse],
    UploadableStorageIO[InstanceSelector, InstanceResponse, InstanceRequest],
):
    """This class provides functionality to interact with instances in Cognite Data Fusion (CDF).

    It is used to download, upload, and purge instances, as well as spaces,views, and containers related to instances.

    Args:
        client (ToolkitClient): An instance of ToolkitClient to interact with the CDF API.
        remove_existing_version (bool): Whether to remove existing versions of instances during upload.
            Default is True. Existing versions are used to safeguard against accidental overwrites.

    """

    KIND = "Instances"
    DISPLAY_NAME = "Instances"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = "/models/instances"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = MappingProxyType({"autoCreateDirectRelations": True})
    BASE_SELECTOR = InstanceSelector

    def __init__(self, client: ToolkitClient, remove_existing_version: bool = True) -> None:
        super().__init__(client)
        self._remove_existing_version = remove_existing_version
        # Cache for view to read-only properties mapping
        self._view_readonly_properties_cache: dict[ViewReference, set[str]] = {}
        self._view_crud = ViewCRUD.create_loader(self.client)

    def as_id(self, item: InstanceResponse) -> str:
        return f"{item.space}:{item.external_id}"

    @staticmethod
    def _build_instance_filter(selector: InstanceViewSelector | InstanceSpaceSelector) -> InstanceFilter:
        """Build an InstanceFilter from a selector.

        Args:
            selector: The selector to build the filter from.

        Returns:
            An InstanceFilter for the toolkit instances API.
        """
        source: TypedViewReference | None = None
        space: list[str] | None = None

        if isinstance(selector, InstanceViewSelector):
            source = TypedViewReference(
                space=selector.view.space,
                external_id=selector.view.external_id,
                version=selector.view.version or "",
            )
            if selector.instance_spaces:
                space = list(selector.instance_spaces)
        elif isinstance(selector, InstanceSpaceSelector):
            space = [selector.instance_space]
            if selector.view and selector.view.version:
                source = TypedViewReference(
                    space=selector.view.space,
                    external_id=selector.view.external_id,
                    version=selector.view.version,
                )

        return InstanceFilter(
            instance_type=selector.instance_type,
            source=source,
            space=space,
        )

    def _filter_readonly_properties(self, instance: InstanceRequest) -> None:
        """Filter out read-only properties from the instance.

        Warnings: This mutates the instance in-place.

        This is as of 17/02/26, the path, root, and pathLastUpdatedTime time in the CogniteAsset container,
        and isUploaded and uploadedTime in the CogniteFile container.
        """
        if not instance.sources:
            return

        for source in instance.sources:
            if source.properties is None:
                continue
            readonly_properties: set[str] = set()
            if isinstance(source.source, ViewReference):
                if source.source not in self._view_readonly_properties_cache:
                    self._view_readonly_properties_cache[source.source] = self._view_crud.get_readonly_properties(
                        source.source
                    )
                readonly_properties = self._view_readonly_properties_cache[source.source]
            elif isinstance(source.source, ContainerReference):
                if source.source.as_tuple() in constants.READONLY_CONTAINER_PROPERTIES:
                    readonly_properties = constants.READONLY_CONTAINER_PROPERTIES[source.source.as_tuple()]
            if not readonly_properties:
                continue
            source.properties = {k: v for k, v in source.properties.items() if k not in readonly_properties}

    def stream_data(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[Page]:
        if isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            instance_filter = self._build_instance_filter(selector)
            total = 0
            cursor: str | None = None
            while cursor is not None or total == 0:
                page_limit = min(self.CHUNK_SIZE, limit - total) if limit is not None else self.CHUNK_SIZE
                page = self.client.tool.instances.paginate(instance_filter, limit=page_limit, cursor=cursor)
                total += len(page.items)
                if page:
                    yield Page(worker_id="main", items=page.items, next_cursor=page.next_cursor)
                if page.next_cursor is None or (limit is not None and total >= limit):
                    break
                cursor = page.next_cursor
        elif isinstance(selector, InstanceFileSelector):
            for chunk in chunker_sequence(selector.ids, self.CHUNK_SIZE):
                yield Page(worker_id="main", items=self.client.tool.instances.retrieve(chunk))
        else:
            raise NotImplementedError()

    def download_ids(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[Sequence[InstanceId]]:
        # Todo: Switch to use pydantic classes once purge has been updated.
        if isinstance(selector, InstanceFileSelector) and selector.validate_instance is False:
            instances_to_yield = selector.instance_ids
            if limit is not None:
                instances_to_yield = instances_to_yield[:limit]
            yield from chunker_sequence(instances_to_yield, self.CHUNK_SIZE)
        else:
            yield from (
                [
                    NodeId(space=instance.space, external_id=instance.external_id)
                    if instance.instance_type == "node"
                    else EdgeId(space=instance.space, external_id=instance.external_id)
                    for instance in chunk.items
                ]
                for chunk in self.stream_data(selector, limit)
            )

    def count(self, selector: InstanceSelector) -> int | None:
        if isinstance(selector, InstanceViewSelector) or (
            isinstance(selector, InstanceSpaceSelector) and selector.view
        ):
            view_id = cast(SelectedView, selector.view)
            result = self.client.data_modeling.instances.aggregate(
                view=ViewId(space=view_id.space, external_id=view_id.external_id, version=view_id.version),
                aggregates=Count("externalId"),
                instance_type=selector.instance_type,
                space=selector.get_instance_spaces(),
            )
            return int(result.value or 0)
        elif isinstance(selector, InstanceSpaceSelector):
            statistics = self.client.data_modeling.statistics.spaces.retrieve(space=selector.instance_space)
            if statistics is None:
                return None
            if selector.instance_type == "node":
                return statistics.nodes
            elif selector.instance_type == "edge":
                return statistics.edges
            # This should never happen due to validation in the selector.
            raise ValueError(f"Unknown instance type {selector.instance_type!r}")
        elif isinstance(selector, InstanceFileSelector):
            return len(selector.items)
        raise NotImplementedError()

    def data_to_json_chunk(
        self, data_chunk: Sequence[InstanceResponse], selector: InstanceSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        return [instance.as_request_resource().dump() for instance in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceRequest:
        item_to_load = item_json.copy()
        if self._remove_existing_version and "existingVersion" in item_to_load:
            del item_to_load["existingVersion"]
        instance = InstanceRequestAdapter.validate_python(item_to_load)
        self._filter_readonly_properties(instance)
        return instance

    def configurations(self, selector: InstanceSelector) -> Iterable[StorageIOConfig]:
        if not isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            return
        spaces = list(set((selector.get_instance_spaces() or []) + (selector.get_schema_spaces() or [])))
        if not spaces:
            return
        space_crud = SpaceCRUD.create_loader(self.client)
        retrieved_spaces = space_crud.retrieve([SpaceReference(space=space) for space in spaces])
        retrieved_spaces = [space for space in retrieved_spaces if not space.is_global]
        if not retrieved_spaces:
            return
        for space in retrieved_spaces:
            yield StorageIOConfig(
                kind=SpaceCRUD.kind,
                folder_name=SpaceCRUD.folder_name,
                value=space_crud.dump_resource(space),
                filename=sanitize_filename(space.space),
            )
        if not selector.view:
            return
        view_crud = ViewCRUD(self.client, None, None, topological_sort_implements=True)
        views = self.client.tool.views.retrieve([selector.view.as_id()], include_inherited_properties=False)
        views = [view for view in views if not view.is_global]
        if not views:
            return
        for view in views:
            filename = f"{view.space}_{view.external_id}"
            if view.version is not None:
                filename += f"_{view.version}"
            yield StorageIOConfig(
                kind=ViewCRUD.kind,
                folder_name=ViewCRUD.folder_name,
                value=view_crud.dump_resource(view),
                filename=sanitize_filename(filename),
            )
        container_ids = list({container for view in views for container in view.mapped_containers})
        if not container_ids:
            return
        container_crud = ContainerCRUD.create_loader(self.client)
        containers = container_crud.retrieve(container_ids)
        containers = [container for container in containers if not container.is_global]
        if not containers:
            return
        for container in containers:
            yield StorageIOConfig(
                kind=ContainerCRUD.kind,
                folder_name=ContainerCRUD.folder_name,
                value=container_crud.dump_resource(container),
                filename=sanitize_filename(f"{container.space}_{container.external_id}"),
            )



from collections import defaultdict
from pathlib import Path
from typing import Any

from cognite_toolkit._cdf_tk.commands.build_v2.data_classes import (
    AbsoluteDirPath,
    InsightList,
    ModelSyntaxError,
    ModuleSource,
    RelativeDirPath,
)
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import BuildVariable
from cognite_toolkit._cdf_tk.constants import EXCL_FILES
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA


class ModuleSourceParser:
    MODULE_ERROR_CODE = "MOD_001"
    VARIABLE_ERROR_CODE = "CONFIG_VARIABLE_001"

    def __init__(self, selected_modules: set[RelativeDirPath | str], organization_dir: AbsoluteDirPath) -> None:
        self.selected_modules = selected_modules
        self.organization_dir = organization_dir
        self.errors = InsightList()

    def parse(self, yaml_files: list[RelativeDirPath], variables: dict[str, Any]) -> list[ModuleSource]:
        files_by_module, orphans = self._find_modules(yaml_files)
        errors = self._validate_modules(list(files_by_module.keys()), self.selected_modules, orphans)
        if errors:
            self.errors.extend(errors)
            return []
        selected_modules = self._select_modules(files_by_module, self.selected_modules)
        build_variables, errors = self._parse_variables(variables, set(files_by_module.keys()), set(selected_modules))
        if errors:
            self.errors.extend(errors)
            return []
        module_sources: list[ModuleSource] = []
        for module in selected_modules:
            source = ModuleSource(
                path=self.organization_dir / module,
                id=module,
                resource_files=[self.organization_dir / resource_file for resource_file in files_by_module[module]],
            )
            module_build_variables = build_variables.get(module, [])
            if module_build_variables:
                for iteration, module_variable in enumerate(module_build_variables, start=1):
                    module_sources.append(
                        source.model_copy(
                            update={
                                "variables": module_variable,
                                "iteration": iteration,
                            }
                        )
                    )
            else:
                module_sources.append(source)
        return module_sources

    @classmethod
    def _find_modules(
        cls, yaml_files: list[RelativeDirPath]
    ) -> tuple[dict[RelativeDirPath, list[RelativeDirPath]], list[RelativeDirPath]]:
        """Organizes YAML files by their module (top-level folder in the modules directory)."""
        files_by_module: dict[RelativeDirPath, list[RelativeDirPath]] = defaultdict(list)
        orphan_files: list[RelativeDirPath] = []
        for yaml_file in yaml_files:
            if yaml_file.name in EXCL_FILES:
                continue
            module_path = cls._get_module_path_from_resource_file_path(yaml_file)
            if module_path:
                files_by_module[module_path].append(yaml_file)
            else:
                orphan_files.append(yaml_file)
        return dict(files_by_module), orphan_files

    @staticmethod
    def _get_module_path_from_resource_file_path(resource_file: Path) -> Path | None:
        for parent in resource_file.parents:
            if parent.name in CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA:
                return parent.parent
        return None

    @classmethod
    def _validate_modules(
        cls,
        module_paths: list[RelativeDirPath],
        selection: set[RelativeDirPath | str],
        orphan_files: list[RelativeDirPath],
    ) -> list[ModelSyntaxError]:
        # Todo: Check the following
        # 1) If any module is inside another module. -> Error.
        # 2) If any selection by name is ambiguous (matches multiple found modules) -> Error.
        # 3) Any selected module by name that does not match any found module -> Error.
        # 4) If any orphan files that have a resource kind set are found -> Error.
        return []

    @classmethod
    def _select_modules(
        cls, files_by_module: dict[RelativeDirPath, list[RelativeDirPath]], selection: set[RelativeDirPath | str]
    ) -> list[RelativeDirPath]:
        return [
            module_path
            for module_path in files_by_module.keys()
            if module_path in selection
            or module_path.name in selection
            or any(parent in selection for parent in module_path.parents)
        ]

    @classmethod
    def _parse_variables(
        cls,
        variables: dict[str, Any],
        available_modules: set[RelativeDirPath],
        selected_modules: set[RelativeDirPath],
    ) -> tuple[dict[RelativeDirPath, list[list[BuildVariable]]], list[ModelSyntaxError]]:
        return {}, []



from pydantic import BaseModel, ConfigDict, DirectoryPath, Field

from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource

from ._insights import InsightList
from ._types import AbsoluteFilePath, RelativeDirPath


class BuildVariable(BaseModel): ...


class ModuleSource(BaseModel):
    """Class used to describe source for module"""

    path: DirectoryPath = Field(description="Path to the module directory. Can be relative or absolute.")
    id: RelativeDirPath = Field(description="Relative path to the organization directory.")
    resource_files: list[AbsoluteFilePath] = Field(default_factory=list)
    variables: list[BuildVariable] = Field(default_factory=list)
    iteration: int = 0

    @property
    def name(self) -> str:
        return self.path.name


class ResourceType(BaseModel):
    model_config = ConfigDict(frozen=True)

    resource_folder: str
    kind: str


class Module(BaseModel):
    """Class used to store module in-memory"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    source: ModuleSource

    resources_by_type: dict[ResourceType, list[ToolkitResource]]
    insights: InsightList = Field(default_factory=InsightList)

    @property
    def is_success(self) -> bool:
        return not self.insights.has_model_syntax_errors



import os
import sys
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path

from pydantic import JsonValue, ValidationError
from rich.console import Console
from rich.panel import Panel

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.build_v2._module_source_parser import ModuleSourceParser
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes import (
    BuildFiles,
    BuildFolder,
    BuildParameters,
    BuiltModule,
    ConfigYAML,
    InsightList,
    Module,
    ModuleSource,
    RelativeDirPath,
    ResourceType,
    ValidationType,
)
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import ModelSyntaxError, Recommendation
from cognite_toolkit._cdf_tk.constants import HINT_LEAD_TEXT, MODULES
from cognite_toolkit._cdf_tk.cruds import RESOURCE_CRUD_BY_FOLDER_NAME
from cognite_toolkit._cdf_tk.cruds._resource_cruds.datamodel import DataModelCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError, ToolkitNotADirectoryError, ToolkitValueError
from cognite_toolkit._cdf_tk.resource_classes import ToolkitResource
from cognite_toolkit._cdf_tk.utils import read_yaml_file, safe_write
from cognite_toolkit._cdf_tk.utils.file import relative_to_if_possible, yaml_safe_dump
from cognite_toolkit._cdf_tk.validation import humanize_validation_error


class BuildV2Command(ToolkitCommand):
    def build(self, parameters: BuildParameters, client: ToolkitClient | None = None) -> BuildFolder:
        console = client.console if client else Console()

        self._validate_build_parameters(parameters, console, sys.argv)
        build_files = self._read_file_system(parameters)
        module_sources = self._parse_module_sources(build_files)

        build_folder = self._build_modules(module_sources, parameters.build_dir)

        # Todo: Some mixpanel tracking.
        # Can be parallelized with number of plugins.
        # Neat is done inside the global validation.
        self._global_validation(build_folder, client)

        self._write_results(build_folder)
        return build_folder

    @classmethod
    def _validate_build_parameters(cls, parameters: BuildParameters, console: Console, user_args: list[str]) -> None:
        """Checks that the user has the correct folders set up and that the config file (if provided) exists."""

        # Set up the variables
        organization_dir = parameters.organization_dir
        organization_dir_display = relative_to_if_possible(organization_dir)
        config_yaml_path: Path | None = None
        if parameters.config_yaml_name:
            config_yaml_path = organization_dir / ConfigYAML.get_filename(parameters.config_yaml_name)
            content = f"  ┣ {MODULES}/\n  ┗ {config_yaml_path.name}\n"
        else:
            content = f"  ┗ {MODULES}\n"
        expected_panel = Panel(
            f"Toolkit expects the following structure:\n{organization_dir_display.as_posix()!r}/\n{content}",
            expand=False,
        )
        module_directory = parameters.modules_directory

        # Execute the checks.
        if module_directory.exists() and (config_yaml_path is None or config_yaml_path.exists()):
            # All good.
            return

        console.print(expected_panel)
        if not organization_dir.exists():
            raise ToolkitNotADirectoryError(f"Organization directory '{organization_dir_display.as_posix()}' not found")
        elif module_directory.exists() and config_yaml_path is not None and not config_yaml_path.exists():
            raise ToolkitFileNotFoundError(
                f"Config YAML file '{relative_to_if_possible(config_yaml_path).as_posix()}' not found"
            )
        elif not module_directory.exists():
            # This is likely the most common error. The user pass in Path.cwd() as the organization directory,
            # but the actual organization directory is a subdirectory.
            candidate_org = next(
                (subdir for subdir in Path.cwd().iterdir() if subdir.is_dir() and (subdir / MODULES).exists()), None
            )
            if candidate_org:
                display_path = relative_to_if_possible(candidate_org)
                suggested_command = cls._create_suggested_command(display_path, user_args)
                console.print(
                    f"\n{HINT_LEAD_TEXT}Did [red]you[/red] mean to use the command {suggested_command}?\n",
                    markup=True,
                )
                cdf_toml = CDFToml.load()
                if not cdf_toml.cdf.has_user_set_default_org:
                    console.print(
                        f"{HINT_LEAD_TEXT} You can specify a 'default_organization_dir = ...' in the 'cdf' section of your "
                        f"'{CDFToml.file_name}' file to avoid using the -o/--organization-dir argument",
                        markup=True,
                    )
            raise ToolkitNotADirectoryError(
                f"Could not find the modules directory at '{relative_to_if_possible(module_directory).as_posix()}'"
            )
        else:
            raise NotImplementedError("Unhandled case. Please report this.")

    @classmethod
    def _create_suggested_command(cls, display_path: Path, user_args: list[str]) -> str:
        suggestion = ["cdf"]
        skip_next = False
        found = False
        for arg in user_args[1:]:
            if skip_next:
                skip_next = False
                continue

            arg_name = arg.split("=")[0]
            if arg_name in ("-o", "--organization-dir"):
                suggestion.append(f"{arg_name} {display_path}")
                found = True
                if "=" not in arg:
                    skip_next = True
                continue
            suggestion.append(arg)
        if not found:
            suggestion.append(f"-o {display_path}")
        return f"'{' '.join(suggestion)}'"

    def _parse_module_sources(self, build: BuildFiles) -> list[ModuleSource]:
        parser = ModuleSourceParser(
            build.selected_modules,
            build.organization_dir,
        )
        module_sources = parser.parse(build.yaml_files, build.variables)
        if parser.errors:
            # Todo: Nicer way of formatting errors.
            raise ToolkitValueError(
                "Errors encountered while parsing modules:\n" + "\n".join(f"- {error!s}" for error in parser.errors)
            )
        return module_sources

    @classmethod
    def _read_file_system(cls, parameters: BuildParameters) -> BuildFiles:
        """Reads the file system to find the YAML files to build along with config.<name>.yaml if it exists."""
        selected: set[RelativeDirPath | str] = {
            parameters.modules_directory.relative_to(parameters.organization_dir)
        }  # Default to everything under modules.
        variables: dict[str, JsonValue] = {}
        cdf_project: str = os.environ.get("CDF_PROJECT", "UNKNOWN")
        validation_type: ValidationType = "prod"
        if parameters.user_selected_modules:
            selected, errors = cls._parse_user_selection(parameters.user_selected_modules, parameters.organization_dir)
            if errors:
                raise ToolkitValueError("Invalid module selection:\n" + "\n".join(f"- {error}" for error in errors))

        if parameters.config_yaml_name:
            try:
                config = ConfigYAML.from_yaml_file(
                    ConfigYAML.get_filepath(parameters.organization_dir, parameters.config_yaml_name)
                )
            except ValidationError as e:
                errors = humanize_validation_error(e)
                raise ToolkitValueError(
                    f"Config YAML file '{parameters.config_yaml_name}' is invalid:\n{'- '.join(errors)}"
                ) from e
            if not parameters.user_selected_modules and config.environment.selected:
                selected, errors = cls._parse_user_selection(config.environment.selected, parameters.organization_dir)
                if errors:
                    raise ToolkitValueError("Invalid module selection:\n" + "\n".join(f"- {error}" for error in errors))
            variables = config.variables or {}
            cdf_project = config.environment.project
            validation_type = config.environment.validation_type

        # Todo optimize by only searching for yaml files in the selected modules paths if selection is provided.
        yaml_files = [
            yaml_file.relative_to(parameters.organization_dir)
            for yaml_file in parameters.modules_directory.rglob("*.y*ml")
        ]
        return BuildFiles(
            yaml_files=yaml_files,
            selected_modules=selected,
            variables=variables,
            validation_type=validation_type,
            cdf_project=cdf_project,
            organization_dir=parameters.organization_dir,
        )

    @classmethod
    def _parse_user_selection(
        cls, user_selected_modules: list[str], organization_dir: Path
    ) -> tuple[set[RelativeDirPath | str], list[str]]:
        selected: set[RelativeDirPath | str] = set()
        errors: list[str] = []
        for item in user_selected_modules:
            if "/" not in item:
                # Module name provided
                selected.add(item)
                continue

            item_path = Path(item)
            if not Path(item).resolve().is_relative_to(organization_dir):
                errors.append(f"Selected module path {item_path.as_posix()!r} is not under the organization directory")
                continue

            if item_path.is_absolute():
                errors.append(
                    f"Selected module path {item_path.as_posix()!r} should be relative to the organization directory"
                )
                continue
            if not (organization_dir / item_path).exists():
                errors.append(
                    f"Selected module path {item_path.as_posix()!r} does not exist under the organization directory"
                )
                continue
            if not item_path.is_dir():
                errors.append(f"Selected module path {item_path.as_posix()!r} is not a directory")
                continue
            selected.add(item_path)
        return selected, errors

    def _build_modules(
        self, module_sources: Iterable[ModuleSource], build_dir: Path, max_workers: int = 1
    ) -> BuildFolder:
        folder: BuildFolder = BuildFolder(path=build_dir)

        for source in module_sources:
            # Inside this loop, do not raise exceptions.
            module = self._import_module(source)  # Syntax validation

            # init built_module, add syntax errors if there are any,
            built_module = BuiltModule(source=module.source, insights=module.insights)

            if module.is_success:
                built_module.insights.extend(self._local_validation(module))
                built_module.built_files = self._export_module(module, build_dir)

            folder.built_modules.append(built_module)

        return folder

    def _import_module(self, module_source: ModuleSource) -> Module:
        insights: InsightList = InsightList()
        resource_folder_paths = [
            resource_path for resource_path in module_source.path.iterdir() if resource_path.is_dir()
        ]

        resource_by_type: dict[ResourceType, list[ToolkitResource]] = defaultdict(list)
        for resource_folder_path in resource_folder_paths:
            crud_classes = RESOURCE_CRUD_BY_FOLDER_NAME.get(resource_folder_path.name)
            if not crud_classes:
                # This is handled in the module parsing phase.
                continue
            for crud_class in crud_classes:
                resource_type = ResourceType(resource_folder=resource_folder_path.name, kind=crud_class.kind)
                resource_files = list(resource_folder_path.rglob(f"*.{crud_class.kind}.y*ml"))
                for resource_file in resource_files:
                    # Todo: Create a classmethod for ToolkitResource
                    # Todo; Handle lists of resources in a single file
                    try:
                        resource = crud_class.yaml_cls.model_validate(read_yaml_file(resource_file))
                        resource_by_type[resource_type].append(resource)
                    except ValidationError as e:
                        insights.extend(self._create_syntax_errors(resource_type, e))

        return Module(source=module_source, resources_by_type=resource_by_type, insights=insights)

    def _export_module(self, module: Module, build_dir: Path) -> list[Path]:
        build_dir.mkdir(parents=True, exist_ok=True)

        built_files: list[Path] = []
        for resource_type, resources in module.resources_by_type.items():
            folder = build_dir / resource_type.resource_folder
            folder.mkdir(parents=True, exist_ok=True)
            for index, resource in enumerate(resources, start=1):
                resource_file = folder / f"resource_{index}.{resource_type.kind}.yaml"
                # Todo Move into Toolkit resource.
                safe_write(resource_file, yaml_safe_dump(resource.model_dump(by_alias=True, exclude_unset=True)))
                built_files.append(resource_file)

        # Todo: Store source path, source hash, ID, and so on for build_linage
        return built_files

    @classmethod
    def _create_syntax_errors(cls, resource_type: ResourceType, error: ValidationError) -> Iterable[ModelSyntaxError]:
        # TODO: should be extended with humanizing of errors, this is a quick solution.

        for error_details in error.errors(include_input=True, include_url=False):
            message = error_details.get("msg", "Unknown syntax error")
            yield ModelSyntaxError(code=f"{resource_type}-SYNTAX-ERROR", message=message)

    def _local_validation(self, module: Module) -> InsightList:
        """This validation is performed locally per entire module"""
        return InsightList()

    def _global_validation(self, build_folder: BuildFolder, client: ToolkitClient | None) -> None:
        """This validation is performed per resource type and not per individual resource and against CDF
        for all modules. This validation will leverage external plugins such as NEAT.
        """

        # Can be parallelized if needed
        for built_module in build_folder.built_modules:
            if not built_module.is_success:
                continue

            if files_by_resource_type := built_module.resource_by_type.get(DataModelCRUD.folder_name):
                built_module.insights.extend(self._validate_with_neat(files_by_resource_type, client))

    def _validate_with_neat(
        self, files_by_resource_type: dict[str, list[Path]], client: ToolkitClient | None
    ) -> InsightList:
        """Placeholder for NEAT validation."""

        insights = InsightList()

        try:
            from cognite.neat._issues import Recommendation as NeatRecommendation

            neat_insight = NeatRecommendation(
                message="Good job! You are using Neat!",
                code="NEAT-USER",
            )
            insights.append(Recommendation.model_validate(neat_insight.model_dump()))
        except ImportError:
            local_insight = Recommendation(
                message="It is always a good idea to use Neat!",
                code="NEAT-EVANGELISM",
                fix="pip install cognite-toolkit[v08]",
            )
            insights.append(local_insight)

        return insights

    def _write_results(self, build_folder: BuildFolder) -> None:
        return None



from cognite.client import CogniteClient

from cognite_toolkit._cdf_tk.client.config import ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, RequestMessage
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.project import ProjectStatusList
from cognite_toolkit._cdf_tk.client.resource_classes.project import OrganizationResponse


class ProjectAPI:
    def __init__(self, config: ToolkitClientConfig, cognite_client: CogniteClient) -> None:
        self._config = config
        self._cognite_client = cognite_client
        self._http_client = HTTPClient(config, split_items_status_codes=set())

    def status(self) -> ProjectStatusList:
        """Retrieve information about the current project."""
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=f"{self._config.base_url}/api/v1/projects",
                method="GET",
                parameters={"withDataModelingStatus": True},
            )
        )
        success = response.get_success_or_raise()
        return ProjectStatusList._load(success.body_json["items"], cognite_client=self._cognite_client)

    def organization(self) -> OrganizationResponse:
        """Retrieve information about the organization of the current project."""
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._config.base_api_url,
                method="GET",
            )
        )
        success = response.get_success_or_raise()
        return OrganizationResponse.model_validate_json(success.body)



from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import ExternalId


class Label(BaseModelObject):
    external_id: str
    name: str
    description: str | None = None
    data_set_id: int | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class LabelRequest(Label, RequestResource): ...


class LabelResponse(Label, ResponseResource[LabelRequest]):
    created_time: int

    def as_request_resource(self) -> LabelRequest:
        return LabelRequest.model_validate(self.dump(), extra="ignore")



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_destination import (
    HostedExtractorDestinationRequest,
    HostedExtractorDestinationResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class HostedExtractorDestinationsAPI(
    CDFResourceAPI[ExternalId, HostedExtractorDestinationRequest, HostedExtractorDestinationResponse]
):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/hostedextractors/destinations", item_limit=10),
                "retrieve": Endpoint(method="POST", path="/hostedextractors/destinations/byids", item_limit=100),
                "update": Endpoint(method="POST", path="/hostedextractors/destinations/update", item_limit=10),
                "delete": Endpoint(method="POST", path="/hostedextractors/destinations/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/hostedextractors/destinations", item_limit=100),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[HostedExtractorDestinationResponse]:
        return PagedResponse[HostedExtractorDestinationResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[HostedExtractorDestinationRequest]) -> list[HostedExtractorDestinationResponse]:
        """Create hosted extractor destinations in CDF.

        Args:
            items: List of destination request objects to create.
        Returns:
            List of created destination response objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self,
        items: Sequence[ExternalId],
        ignore_unknown_ids: bool = False,
    ) -> list[HostedExtractorDestinationResponse]:
        """Retrieve hosted extractor destinations from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved destination response objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[HostedExtractorDestinationRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[HostedExtractorDestinationResponse]:
        """Update hosted extractor destinations in CDF.

        Args:
            items: List of destination request objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated destination response objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False, force: bool | None = None) -> None:
        """Delete hosted extractor destinations from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
            force: Delete any jobs associated with each item.
        """
        extra_body: dict[str, Any] = {"ignoreUnknownIds": ignore_unknown_ids}
        if force is not None:
            extra_body["force"] = force

        self._request_no_response(items, "delete", extra_body=extra_body)

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[HostedExtractorDestinationResponse]:
        """Iterate over hosted extractor destinations in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of destination response objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(
        self,
        limit: int = 100,
    ) -> Iterable[list[HostedExtractorDestinationResponse]]:
        """Iterate over hosted extractor destinations in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of destination response objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 100) -> list[HostedExtractorDestinationResponse]:
        """List all hosted extractor destinations in CDF.

        Args:
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of destination response objects.
        """
        return self._list(limit=limit)



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.api.transformation_notifications import TransformationNotificationsAPI
from cognite_toolkit._cdf_tk.client.api.transformation_schedules import TransformationSchedulesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import TransformationFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.transformation import TransformationRequest, TransformationResponse


class TransformationsAPI(CDFResourceAPI[InternalOrExternalId, TransformationRequest, TransformationResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/transformations", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/transformations/byids", item_limit=1000),
                "update": Endpoint(method="POST", path="/transformations/update", item_limit=1000),
                "delete": Endpoint(method="POST", path="/transformations/delete", item_limit=1000),
                "list": Endpoint(method="POST", path="/transformations/filter", item_limit=1000),
            },
        )
        self.schedules = TransformationSchedulesAPI(http_client)
        self.notifications = TransformationNotificationsAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[TransformationResponse]:
        return PagedResponse[TransformationResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[TransformationRequest]) -> list[TransformationResponse]:
        """Create transformations in CDF.

        Args:
            items: List of TransformationRequest objects to create.
        Returns:
            List of created TransformationResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False, with_job_details: bool = False
    ) -> list[TransformationResponse]:
        """Retrieve transformations from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
            with_job_details: Whether the transformations will be returned with running job and last created job details.
        Returns:
            List of retrieved TransformationResponse objects.
        """
        return self._request_item_response(
            items,
            method="retrieve",
            extra_body={"ignoreUnknownIds": ignore_unknown_ids, "withJobDetails": with_job_details},
        )

    def update(
        self, items: Sequence[TransformationRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[TransformationResponse]:
        """Update transformations in CDF.

        Args:
            items: List of TransformationRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated TransformationResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete transformations from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: TransformationFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[TransformationResponse]:
        """Iterate over all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of TransformationResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter.model_dump(by_alias=True) if filter else {}},
        )

    def iterate(
        self,
        filter: TransformationFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[TransformationResponse]]:
        """Iterate over all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of TransformationResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={"filter": filter.model_dump(by_alias=True) if filter else {}},
        )

    def list(
        self,
        filter: TransformationFilter | None = None,
        limit: int | None = 100,
    ) -> list[TransformationResponse]:
        """List all transformations in CDF.

        Args:
            filter: TransformationFilter object to filter transformations.
            limit: Maximum number of items to return.

        Returns:
            List of TransformationResponse objects.
        """
        return self._list(limit=limit, body={"filter": filter.model_dump(by_alias=True) if filter else {}})



from typing import ClassVar

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class DataSet(BaseModelObject):
    external_id: str | None = None
    name: str | None = None
    description: str | None = None
    metadata: dict[str, str] | None = None
    write_protected: bool = False

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert DataSet to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class DataSetRequest(DataSet, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata"})


class DataSetResponse(DataSet, ResponseResource[DataSetRequest]):
    id: int
    created_time: int
    last_updated_time: int

    def as_request_resource(self) -> DataSetRequest:
        return DataSetRequest.model_validate(self.dump(), extra="ignore")
